/*
 *  callbacks.c - functions to handle GUI events.
 *	part of galculator
 *  	(c) 2002-2003 Simon Floery (simon.floery@gmx.at)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "calc_basic.h"
#include "galculator.h"
#include "math_functions.h"
#include "general_functions.h"
#include "display.h"
#include "config_file.h"

#include "callbacks.h"
#include "interface.h"
#include "support.h"

/* trigonometric functions have to pay attention to current angle base (rad, deg or grad).
 * nevertheless they are stored together with the other functions (and served by the same
 * callback. To distinguish between normal and trigonometric functions, the latter ones are
 * stored at the beginning of the function_list. The constant TRIGO_FUNC_STOP is the
 * index where the normal functions start.
 */

#define TRIGO_FUNC_STOP	3

s_function_list function_list[] = {\
	{"sin", {sin, asin, sinh, sin}},\
	{"cos", {cos, acos, cosh, cos}},\
	{"tan", {tan, atan, tanh, tan}},\
	{"log", {log10, pow10y, log10, log10}},\
	{"ln", {log, exp, log, log}},\
	{"1/x", {reciprocal, idx, reciprocal, reciprocal}},\
	{"x^2", {powx2, sqrt, powx2, powx2}},\
	{"sqrt", {sqrt, powx2, sqrt, sqrt}},\
	{"n!", {factorial, factorial, factorial, factorial}},\
	{"CMP", {cmp, cmp, cmp, cmp}},\
	{NULL}\
};

s_constant_list constant_list[] = {\
	{"e", G_E},\
	{"Pi", G_PI},\
	{NULL}\
};

s_gfunc_list gfunc_list[] = {\
	{"+/-", display_result_toggle_sign},\
	{"<-", display_result_backspace},\
	{"EE", display_append_e},\
	{"C", clear},\
	{"AC", all_clear},\
	{"MS", memory_save},\
	{"MR", memory_read},\
	{"M+", memory_add2mem},\
	{NULL}\
};

s_operation_map operation_map[] = {\
	{"x^y", '^'},\
	{"LSH", '<'},\
	{"MOD", '%'},\
	{"AND", '&'},\
	{"OR", '|'},\
	{"XOR", 'x'},\
	{"ENT", '='},\
	{NULL}\
};

static GtkWidget		*about_dialog, *prefs_dialog, *font_dialog, *color_dialog;
extern GtkWidget		*main_window;
gboolean				rpn_have_result=FALSE, allow_arith_op=TRUE;
static GtkToggleButton 	*inv_button=NULL, *hyp_button=NULL;
extern s_preferences	prefs;
extern s_prefs_entry	prefs_list[];
extern s_current_status current_status;

/* File */

void
on_quit_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
	// remember display's value
	if (prefs.rem_value) g_free (prefs.rem_value);
	prefs.rem_value = display_result_get();
	gtk_main_quit();
}

/* Help */

void
on_about_activate                     (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
	GtkLabel	*about_label;
	char		*about_text;
	
	about_dialog = create_about_dialog ();
	gtk_window_set_title ((GtkWindow *)about_dialog, g_strdup_printf ("[about] %s v%s", PROG_NAME, PROG_VERSION));
	about_label = (GtkLabel *)lookup_widget (about_dialog, "about_label");
	gtk_label_set_justify (about_label, GTK_JUSTIFY_CENTER);
	about_text = g_strdup_printf ("<span size=\"x-large\" weight=\"bold\">%s v%s</span>\n\
<span size=\"large\">a GTK 2 based scientific calculator</span>\n\n\
(c) 2002-2003 by Simon Floery (simon.floery@gmx.at)", PROG_NAME, PROG_VERSION);
	gtk_label_set_markup (about_label, \
		about_text);
	gtk_widget_show (about_dialog);
}

/* this callback is called if a button for entering a number is clicked. There are two
 * cases: either starting a new number or appending a digit to the existing number.
 * The decimal point leads to some specialities.
 */

void
on_number_button_clicked               (GtkButton       *button,
                                        gpointer         user_data)
{	
	/* put a ev. result onto the stack */
	/* an alternative idea would be to do this immediately after getting a result
	   in calc_basic's rpn routines or in on_function_button_clicked. but if we
	   use on_function_button_clicked twice, there are two results on the stack
	   where we don't expect the older one to be there. therefore doing it this way
	*/
	if ((current_status.notation == CS_RPN) && (rpn_have_result == TRUE))
	{
		calc_rpn_stack_add (display_result_get_as_double ());
		rpn_have_result = FALSE;
	}
	
	/* append the digit. due to greyed out buttons we only get valid numbers here */
	display_result_add_digit (*(gtk_button_get_label (button)));
}

/* this callback is called if a button for doing one of the arithmetic operations plus, minus, 
 * multiply, divide or power is clicked. it is mainly an interface to the calc_basic code.
 */

void
on_operation_button_clicked            (GtkButton       *button,
                                        gpointer         user_data)
{
	s_calc_token	current_token;
	extern gboolean	calc_entry_start_new;
	int				counter = 0;
	const char		*button_label;
	double			return_value;

	/* current number, get it from the display! */
	current_token.number = display_result_get_as_double ();
	button_label = gtk_button_get_label (button);
	if (strlen (button_label) > 1) {
		while (operation_map[counter].button_label != NULL) 
		{
			if (strcmp (operation_map[counter].button_label, button_label) == 0) break;
			counter++;
		}
		if (operation_map[counter].button_label != NULL) { 
			current_token.operator = operation_map [counter].operation;
			// do inverse left shift is a right shift
			if ((current_token.operator == '<') && \
				(BIT (current_status.fmod, CS_FMOD_FLAG_INV) == 1)) {
					current_token.operator = '>';
					if (inv_button != NULL) gtk_toggle_button_set_active (inv_button, FALSE);
			}
		}
		else error_message ("unknown operation mapping");
	} else current_token.operator = *button_label;
	
	/* notation specific interface code */
	
	if (current_status.notation == CS_PAN)
	{
		/* '(' doesn't pay respect to allow_arith_op but sets it: a+((((((b-...
		 * ')' pays respect to allow_arith_op but doesn't set it: ...+a)))))-...
		 * '=' pays respect to allow_arith_op but doesn't set it: ...+a=
		 * 	(in order to continue with the result on the display)
		 * all other operator pay respect and set allow_arith_op.
		 *
		 * in general, a closing bracket is only useful if there were opening
		 *	brackets.
		 */
		
		if (((current_token.operator == '(') || allow_arith_op) && \
			((current_token.operator != ')') || (display_module_bracket_label_update (GET) > 0))) {
			return_value = calc_tree_add_token (current_token);
			display_result_set_double (return_value);
			display_module_arith_label_update (current_token.operator);
			
			/* setting of allow_arith_op. the missing breaks are wanted */
			switch (current_token.operator) {
				case '=':
					display_module_bracket_label_update (RESET);
					break;
				case ')':
					display_module_bracket_label_update (ONE_LESS);
					break;
				case '(':
					display_module_bracket_label_update (ONE_MORE);
				default:
					allow_arith_op=FALSE;
			}
		}
	}
	else if (current_status.notation == CS_RPN)
	{
		switch (current_token.operator) {
		case '=':
			calc_rpn_stack_add (current_token.number);
			rpn_have_result = FALSE;
			break;
		default:
			display_result_set_double (calc_rpn_stack_operation (current_token));
			rpn_have_result = TRUE;
		}
	}
	else fprintf (stderr, "[%s] on_operation_button_clicked: unknown status. %s\n", PROG_NAME, BUG_REPORT);

	calc_entry_start_new = TRUE;
}

/* this callback is called if a button for a function manipulating the current entry directly
 * is clicked. the array function_list knows the relation between button label and function to
 * call.
 */

void
on_function_button_clicked             (GtkButton       *button,
                                        gpointer         user_data)
{
	int				counter=0;
	extern gboolean	calc_entry_start_new;
	
	/* at first find the corresponding function */
	while (function_list[counter].button_label != NULL) 
	{
		if (strcmp (function_list[counter].button_label, gtk_button_get_label (button)) == 0) break;
		counter++;
	}
	if (function_list[counter].button_label != NULL) 
	{
		/* see header of this file for information about TRIGO_FUNC_STOP
			hyperbolic versions of trigonometric functions doesn't have to pay attention to
			angle base. therefore do it like a normal function
		 */
		if ((counter < TRIGO_FUNC_STOP) && (BIT (current_status.fmod, CS_FMOD_FLAG_HYP) == 0))\
		{
			if (BIT (current_status.fmod, CS_FMOD_FLAG_INV) == 0) 
			{
				display_result_set_double (\
					function_list[counter].func[current_status.fmod](display_result_get_rad_angle()));
			}
			else
			{
				display_result_set_angle (\
					function_list[counter].func[current_status.fmod](display_result_get_as_double()));
			}	
		}
		else
		{
			display_result_set_double (\
				function_list[counter].func[current_status.fmod](display_result_get_as_double()));
		}
		if (inv_button != NULL)
		{
			gtk_toggle_button_set_active (inv_button, FALSE);
			gtk_toggle_button_set_active (hyp_button, FALSE);
		}
		calc_entry_start_new = TRUE;	
		if (current_status.notation == CS_RPN) rpn_have_result = TRUE;
	}
	else error_message ("unknown button <-> function relation");
}

void
on_constant_button_clicked             (GtkButton       *button,
                                        gpointer         user_data)
{
	int		counter=0;
	
	/* at first find the corresponding function */
	while (constant_list[counter].button_label != NULL) 
	{
		if (strcmp (constant_list[counter].button_label, gtk_button_get_label (button)) == 0) break;
		counter++;
	}
	if (constant_list[counter].button_label != NULL) display_result_set_double (constant_list[counter].constant);
	else error_message ("unknown button <-> constant relation");
}

/* tbutton_fmod - these are function modifiers such as INV (inverse) and HYP (hyperbolic).
 * 	as standard math.h doesn't support the inverse of a hyperbolic function, we allow only
 *	one button to be active.
 */

void
on_tbutton_fmod_clicked                (GtkButton       *button,
                                        gpointer         user_data)
{
	if (inv_button == NULL)
	{
		inv_button = (GtkToggleButton *)lookup_widget (((GtkWidget *)button)->parent, "tbutton_inv");
		hyp_button = (GtkToggleButton *)lookup_widget (((GtkWidget *)button)->parent, "tbutton_hyp");
	}
	if (strcmp (gtk_button_get_label (button), "inv") == 0)
	{
		if (BIT (current_status.fmod, CS_FMOD_FLAG_HYP) == 1) gtk_toggle_button_set_active (hyp_button, FALSE);
		current_status.fmod ^= 1 << CS_FMOD_FLAG_INV;
	}
	else if (strcmp (gtk_button_get_label (button), "hyp") == 0)
	{
		if (BIT (current_status.fmod, CS_FMOD_FLAG_INV) == 1) gtk_toggle_button_set_active (inv_button, FALSE);
		current_status.fmod ^= 1 << CS_FMOD_FLAG_HYP;
	}
	else error_message ("unknown function modifier (INV/HYP)");
}

void
on_gfunc_button_clicked                (GtkButton       *button,
                                        gpointer         user_data)
{
	int		counter=0;
	
	while (gfunc_list[counter].button_label != NULL) 
	{
		if (strcmp (gfunc_list[counter].button_label, gtk_button_get_label (button)) == 0) break;
		counter++;
	}
	if (gfunc_list[counter].button_label != NULL) gfunc_list[counter].func();
	else error_message ("unknown button <-> gfunc relation");
}

void
on_about_ok_clicked                    (GtkButton       *button,
                                        gpointer         user_data)
{
	display_update_modules ();
	gtk_widget_destroy (about_dialog);
}

void
on_dec_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
	if (!gtk_check_menu_item_get_active((GtkCheckMenuItem *)menuitem)) return;
	display_change_option (CS_DEC, DISPLAY_OPT_NUMBER);
}


void
on_hex_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
	if (!gtk_check_menu_item_get_active((GtkCheckMenuItem *)menuitem)) return;
	display_change_option (CS_HEX, DISPLAY_OPT_NUMBER);
}


void
on_oct_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
	if (!gtk_check_menu_item_get_active((GtkCheckMenuItem *)menuitem)) return;
	display_change_option (CS_OCT, DISPLAY_OPT_NUMBER);
}


void
on_bin_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
	if (!gtk_check_menu_item_get_active((GtkCheckMenuItem *)menuitem)) return;
	display_change_option (CS_BIN, DISPLAY_OPT_NUMBER);
}


void
on_deg_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
	if (!gtk_check_menu_item_get_active((GtkCheckMenuItem *)menuitem)) return;
	display_change_option (CS_DEG, DISPLAY_OPT_ANGLE);
}


void
on_rad_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
	if (!gtk_check_menu_item_get_active((GtkCheckMenuItem *)menuitem)) return;
	display_change_option (CS_RAD, DISPLAY_OPT_ANGLE);
}


void
on_grad_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
	if (!gtk_check_menu_item_get_active((GtkCheckMenuItem *)menuitem)) return;
	display_change_option (CS_GRAD, DISPLAY_OPT_ANGLE);
}


void
on_ordinary_activate                  (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
	GtkButton	*button;
	
	if (!gtk_check_menu_item_get_active((GtkCheckMenuItem *)menuitem)) return;
	display_change_option (CS_PAN, DISPLAY_OPT_NOTATION);
	calc_rpn_free ();
	all_clear ();
	button = (GtkButton *) lookup_widget (main_window, "button_enter");
	gtk_button_set_label (button, "=");
}


void
on_rpn_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
	GtkButton	*button;
	
	if (!gtk_check_menu_item_get_active((GtkCheckMenuItem *)menuitem)) return;
	display_change_option (CS_RPN, DISPLAY_OPT_NOTATION);
	calc_tree_free ();
	all_clear ();
	button = (GtkButton *) lookup_widget (main_window, "button_enter");
	gtk_button_set_label (button, "ENT");
}

void
on_preferences1_activate               (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
	int			counter=0;
	GtkWidget	*w;
	
	prefs_dialog = create_prefs_dialog ();
	
	gtk_window_set_title ((GtkWindow *)prefs_dialog, \
		g_strdup_printf ("%s Preferences", PROG_NAME));

	// preferences -> gui
	while (prefs_list[counter].key != NULL) {
		if (prefs_list[counter].set_handler != NULL) \
			prefs_list[counter].set_handler (prefs_dialog, \
				prefs_list[counter].widget_name, \
				prefs_list[counter].variable);
		counter++;
	}
	
	w = lookup_widget (prefs_dialog, "prefs_button_font_label");
	gtk_widget_set_sensitive (w, prefs.custom_button_font);
	
	w = lookup_widget (prefs_dialog, "prefs_button_font");
	gtk_widget_set_sensitive (w, prefs.custom_button_font);
		
	gtk_widget_show (prefs_dialog);
}

void
on_prefs_result_font_clicked           (GtkButton       *button,
                                        gpointer         user_data)
{
	font_dialog = show_font_dialog ("Select result font", button);
}


void
on_prefs_result_color_clicked          (GtkButton       *button,
                                        gpointer         user_data)
{
	color_dialog = show_color_dialog ("Select result font color", button);
}


void
on_prefs_mod_font_clicked              (GtkButton       *button,
                                        gpointer         user_data)
{
	font_dialog = show_font_dialog ("Select module font", button);
}


void
on_prefs_act_mod_color_clicked         (GtkButton       *button,
                                        gpointer         user_data)
{
	color_dialog = show_color_dialog ("Select active module color", button);
}


void
on_prefs_inact_mod_color_clicked       (GtkButton       *button,
                                        gpointer         user_data)
{
	color_dialog = show_color_dialog ("Select inactive module color", button);
}


void
on_prefs_bkg_color_clicked             (GtkButton       *button,
                                        gpointer         user_data)
{
	color_dialog = show_color_dialog ("Select background color", button);
}


void
on_prefs_button_font_clicked           (GtkButton       *button,
                                        gpointer         user_data)
{
	font_dialog = show_font_dialog ("Select button font", button);
}

void
on_prefs_apply_clicked                 (GtkButton       *button,
                                        gpointer         user_data)
{
	int 	counter = 0;
	
	while (prefs_list[counter].key != NULL) {
		if (prefs_list[counter].apply_handler != NULL) \
			prefs_list[counter].apply_handler (prefs_dialog, \
				prefs_list[counter].widget_name, \
				prefs_list[counter].variable);
		counter++;
	}

	update_all (main_window, prefs);
}


void
on_prefs_ok_clicked                    (GtkButton       *button,
                                        gpointer         user_data)
{
	/* changes are saved to file when leaving galculator */
	on_prefs_apply_clicked (NULL, NULL);
	gtk_widget_destroy (prefs_dialog);
}


void
on_color_ok_button_clicked             (GtkButton       *button,
                                        gpointer         user_data)
{
	GtkButton		*color_button=NULL;
	const char 		*title;
	GtkWidget		*da;
	GdkColor		color;
	
	title = gtk_window_get_title ((GtkWindow *) color_dialog);
	/* possibilities: background color, result color, act/inactive option color */
	if (strstr (title, " background ") != NULL)
	{
		color_button = (GtkButton *) lookup_widget (prefs_dialog, "prefs_bkg_color");
	}
	else if (strstr (title, " result ") != NULL)
	{
		color_button = (GtkButton *) lookup_widget (prefs_dialog, "prefs_result_color");
	}
	else if (strstr (title, " active ") != NULL)
	{
		color_button = (GtkButton *) lookup_widget (prefs_dialog, "prefs_act_mod_color");
	}
	else if (strstr (title, " inactive ") != NULL)
	{
		color_button = (GtkButton *) lookup_widget (prefs_dialog, "prefs_inact_mod_color");
	}
	else fprintf (stderr, "[%s] Color Dialog (%s) not found. %s\n", PROG_NAME, title, BUG_REPORT);
	
	gtk_color_selection_get_current_color ((GtkColorSelection *)(((GtkColorSelectionDialog *) color_dialog)->colorsel), \
		&color);
	
	da = gtk_bin_get_child ((GtkBin *)gtk_bin_get_child ((GtkBin *) color_button));
	
	gtk_widget_really_modify_fg (da, color);

	gtk_widget_destroy (color_dialog);
}


void
on_font_ok_button_clicked              (GtkButton       *button,
                                        gpointer         user_data)
{
	GtkButton		*font_button=NULL;
	const char 		*title;
	
	title = gtk_window_get_title ((GtkWindow *) font_dialog);
	/*possibilities: result font, option font, button font */
	if (strstr (title, " result ") != NULL)
	{
		font_button = (GtkButton *) lookup_widget (prefs_dialog, "prefs_result_font");
	}
	else if (strstr (title, " module ") != NULL)
	{
		font_button = (GtkButton *) lookup_widget (prefs_dialog, "prefs_mod_font");
	}
	else if (strstr (title, " button ") != NULL)
	{
		font_button = (GtkButton *) lookup_widget (prefs_dialog, "prefs_button_font");
	}
	else fprintf (stderr, "[%s] Font Dialog (%s) not found. %s\n", PROG_NAME, title, BUG_REPORT);
	gtk_button_set_label (font_button, \
		gtk_font_selection_dialog_get_font_name ((GtkFontSelectionDialog *)font_dialog));
	gtk_widget_destroy (font_dialog);
}



void
on_prefs_cancel_clicked                (GtkButton       *button,
                                        gpointer         user_data)
{
	gtk_widget_destroy (prefs_dialog);
}


void
on_color_cancel_button_clicked         (GtkButton       *button,
                                        gpointer         user_data)
{
	gtk_widget_destroy (color_dialog);
}


void
on_font_cancel_button_clicked          (GtkButton       *button,
                                        gpointer         user_data)
{
	gtk_widget_destroy (font_dialog);
}


void
on_show_menubar1_activate              (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
	prefs.show_menu = gtk_check_menu_item_get_active ((GtkCheckMenuItem *) menuitem);;
	set_widget_visibility (main_window, "menubar", prefs.show_menu);
}

void
on_prefs_custom_button_font_toggled    (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
	GtkWidget	*w;
	
	prefs.custom_button_font = gtk_toggle_button_get_active (togglebutton);
	
	w = lookup_widget (prefs_dialog, "prefs_button_font_label");
	gtk_widget_set_sensitive (w, prefs.custom_button_font);
	
	w = lookup_widget (prefs_dialog, "prefs_button_font");
	gtk_widget_set_sensitive (w, prefs.custom_button_font);
}


/* END */
